@@ -11,13 +11,18 @@
 
 #include "first.h"
 
+#include "sys-dirent.h"
+#include "sys-stat.h"
 #include "sys-time.h"
+#include "sys-unistd.h" /* <unistd.h> */
 
 #include "base.h"
 #include "log.h"
 #include "buffer.h"
+#include "chunk.h"
 #include "fdevent.h"
 #include "http_chunk.h"
+#include "http_etag.h"
 #include "http_header.h"
 #include "keyvalue.h"
 #include "response.h"
@@ -29,9 +34,7 @@
 #include <errno.h>
 #include <stdlib.h>
 #include <string.h>
-#include <dirent.h>
 #include <fcntl.h>
-#include <unistd.h>
 
 #ifdef AT_FDCWD
 #ifndef _ATFILE_SOURCE
@@ -39,12 +42,10 @@
 #endif
 #endif
 
-#ifndef _D_EXACT_NAMLEN
-#ifdef _DIRENT_HAVE_D_NAMLEN
-#define _D_EXACT_NAMLEN(d) ((d)->d_namlen)
-#else
-#define _D_EXACT_NAMLEN(d) (strlen ((d)->d_name))
-#endif
+#ifdef _WIN32
+#include <windows.h>
+#include <stringapiset.h>
+#include <stdio.h>      /* FILENAME_MAX */
 #endif
 
 /**
@@ -109,14 +110,16 @@ typedef struct {
 typedef struct {
 	dirls_entry_t **ent;
 	uint32_t used;
-	uint32_t size;
 } dirls_list_t;
 
 #define DIRLIST_ENT_NAME(ent)  ((char*)(ent) + sizeof(dirls_entry_t))
+/* DIRLIST_BLOB_SIZE must be power of 2 for current internal usage */
 #define DIRLIST_BLOB_SIZE      16
 
 typedef struct {
+  #ifndef _WIN32
 	DIR *dp;
+  #endif
 	dirls_list_t dirs;
 	dirls_list_t files;
 	char *path;
@@ -129,22 +132,35 @@ typedef struct {
 	char *jfn;
 	uint32_t jfn_len;
 	plugin_config conf;
+  #ifdef _WIN32
+	HANDLE hFind;
+	WIN32_FIND_DATAW ffd;
+	char fnUTF8[FILENAME_MAX*4+1];
+  #endif
 } handler_ctx;
 
 #define DIRLIST_BATCH 32
 static int dirlist_max_in_progress;
 
 
+__attribute_returns_nonnull__
 static handler_ctx * mod_dirlisting_handler_ctx_init (plugin_data * const p) {
-    handler_ctx *hctx = calloc(1, sizeof(*hctx));
-    force_assert(hctx);
+    handler_ctx *hctx = ck_calloc(1, sizeof(*hctx));
+  #ifdef _WIN32
+    hctx->hFind = INVALID_HANDLE_VALUE;
+  #endif
     memcpy(&hctx->conf, &p->conf, sizeof(plugin_config));
     return hctx;
 }
 
 static void mod_dirlisting_handler_ctx_free (handler_ctx *hctx) {
+  #ifdef _WIN32
+    if (INVALID_HANDLE_VALUE != hctx->hFind)
+        FindClose(hctx->hFind);
+  #else
     if (hctx->dp)
         closedir(hctx->dp);
+  #endif
     if (hctx->files.ent) {
         dirls_entry_t ** const ent = hctx->files.ent;
         for (uint32_t i = 0, used = hctx->files.used; i < used; ++i)
@@ -159,12 +175,12 @@ static void mod_dirlisting_handler_ctx_free (handler_ctx *hctx) {
     }
     if (hctx->jb) {
         chunk_buffer_release(hctx->jb);
+        if (-1 != hctx->jfd)
+            close(hctx->jfd);
         if (hctx->jfn) {
             unlink(hctx->jfn);
             free(hctx->jfn);
         }
-        if (-1 != hctx->jfd)
-            close(hctx->jfd);
     }
     free(hctx->path);
     free(hctx);
@@ -209,8 +225,8 @@ static struct dirlist_cache * mod_dirlisting_parse_cache(server *srv, const arra
         }
     }
 
-    struct dirlist_cache * const cache = calloc(1, sizeof(struct dirlist_cache));
-    force_assert(cache);
+    struct dirlist_cache * const cache =
+      ck_calloc(1, sizeof(struct dirlist_cache));
     cache->max_age = max_age;
     cache->path = path;
     return cache;
@@ -260,7 +276,7 @@ static int mod_dirlisting_exclude(pcre_keyvalue_buffer * const kvb, const char *
 
 
 INIT_FUNC(mod_dirlisting_init) {
-    return calloc(1, sizeof(plugin_data));
+    return ck_calloc(1, sizeof(plugin_data));
 }
 
 FREE_FUNC(mod_dirlisting_free) {
@@ -619,8 +635,10 @@ static void http_list_directory_include_file(request_st * const r, const handler
         buffer_clear(out);
         const int fd = sce->fd;
         ssize_t rd;
+        off_t off = 0;
         char buf[8192];
-        while ((rd = read(fd, buf, sizeof(buf))) > 0) {
+        while ((rd = chunk_file_pread(fd, buf, sizeof(buf), off)) > 0) {
+            off += rd;
             buffer_append_string_encoded(out, buf, (size_t)rd, ENCODING_MINIMAL_XML);
             if (out == tb) {
                 if (0 != chunkqueue_append_mem_to_tempfile(cq,
@@ -640,6 +658,25 @@ static void http_list_directory_include_file(request_st * const r, const handler
     }
 }
 
+static void http_dirlist_link (request_st * const r, const buffer *b, const char *params, size_t plen) {
+    /*(params must be at least ">" to close Link url-reference)*/
+    buffer * const tb = r->tmp_buf;
+    buffer_clear(tb);
+    buffer_append_str3(tb, CONST_STR_LEN("<"), BUF_PTR_LEN(b), params, plen);
+    http_header_response_insert(r, HTTP_HEADER_LINK,
+                                CONST_STR_LEN("Link"),
+                                BUF_PTR_LEN(tb));
+}
+
+static void http_dirlist_auto_layout_early_hints (request_st * const r, const plugin_data * const p) {
+    if (p->conf.external_css)
+        http_dirlist_link(r, p->conf.external_css,
+          CONST_STR_LEN(">; rel=\"preload\"; as=\"style\""));
+    if (p->conf.external_js)
+        http_dirlist_link(r, p->conf.external_js,
+          CONST_STR_LEN(">; rel=\"preload\"; as=\"script\""));
+}
+
 /* portions copied from mod_status
  * modified and specialized for stable dirlist sorting by name */
 static const char js_simple_table_resort[] = \
@@ -986,8 +1023,8 @@ static void http_list_directory_footer(request_st * const r, const handler_ctx *
 
 static int http_open_directory(request_st * const r, handler_ctx * const hctx) {
     const uint32_t dlen = buffer_clen(&r->physical.path);
-#if defined __WIN32
-    hctx->name_max = FILENAME_MAX;
+#ifdef _WIN32
+    hctx->name_max = FILENAME_MAX*4; /*(260 chars * 4 for (max) UTF-8 bytes)*/
 #else
 #ifndef PATH_MAX
 #define PATH_MAX 4096
@@ -995,13 +1032,32 @@ static int http_open_directory(request_st * const r, handler_ctx * const hctx) {
     /* allocate based on PATH_MAX rather than pathconf() to get _PC_NAME_MAX */
     hctx->name_max = PATH_MAX - dlen - 1;
 #endif
-    hctx->path = malloc(dlen + hctx->name_max + 1);
-    force_assert(NULL != hctx->path);
+    hctx->path = ck_malloc(dlen + hctx->name_max + 1);
     memcpy(hctx->path, r->physical.path.ptr, dlen+1);
-  #if defined(HAVE_XATTR) || defined(HAVE_EXTATTR) || !defined(_ATFILE_SOURCE)
+  #if defined(HAVE_XATTR) || defined(HAVE_EXTATTR) \
+   || (!defined(_ATFILE_SOURCE) && !defined(_WIN32))
     hctx->path_file = hctx->path + dlen;
   #endif
 
+  #ifdef _WIN32
+    hctx->path[dlen] = '*';
+    hctx->path[dlen+1] = '\0';
+    WCHAR wbuf[4096];
+    int wlen = MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS,
+                                   hctx->path, dlen+1,
+                                   wbuf, (sizeof(wbuf)/sizeof(*wbuf))-2);
+    if (0 == wlen) return -1;
+    hctx->hFind = FindFirstFileExW(wbuf, FindExInfoBasic, &hctx->ffd,
+                                   FindExSearchNameMatch, NULL,
+                                   FIND_FIRST_EX_LARGE_FETCH);
+    if (INVALID_HANDLE_VALUE == hctx->hFind) {
+        if (GetLastError() != ERROR_FILE_NOT_FOUND) {
+            log_error(r->conf.errh, __FILE__, __LINE__,
+              "FindFirstFileEx failed: %s", r->physical.path.ptr);
+            return -1;
+        }
+    }
+  #else
   #ifndef _ATFILE_SOURCE /*(not using fdopendir unless _ATFILE_SOURCE)*/
     hctx->dfd = -1;
     hctx->dp = opendir(hctx->path);
@@ -1017,34 +1073,50 @@ static int http_open_directory(request_st * const r, handler_ctx * const hctx) {
         }
         return -1;
     }
+  #endif
 
     if (hctx->conf.json) return 0;
 
     dirls_list_t * const dirs = &hctx->dirs;
     dirls_list_t * const files = &hctx->files;
-    dirs->ent   =
-      (dirls_entry_t**) malloc(sizeof(dirls_entry_t*) * DIRLIST_BLOB_SIZE);
-    force_assert(dirs->ent);
-    dirs->size  = DIRLIST_BLOB_SIZE;
+    dirs->ent   = NULL;
     dirs->used  = 0;
-    files->ent  =
-      (dirls_entry_t**) malloc(sizeof(dirls_entry_t*) * DIRLIST_BLOB_SIZE);
-    force_assert(files->ent);
-    files->size = DIRLIST_BLOB_SIZE;
+    files->ent  = NULL;
     files->used = 0;
 
     return 0;
 }
 
 static int http_read_directory(handler_ctx * const p) {
-	struct dirent *dent;
 	const int hide_dotfiles = p->conf.hide_dot_files;
 	const uint32_t name_max = p->name_max;
-	struct stat st;
+  #ifdef _WIN32
+	int count = 0;
+	if (INVALID_HANDLE_VALUE == p->hFind) {
+		/* GetLastError() == ERROR_FILE_NOT_FOUND
+		 * (other errors handled in http_open_directory()) */
+	}
+	else do
+  #else
 	int count = -1;
-	while (++count < DIRLIST_BATCH && (dent = readdir(p->dp)) != NULL) {
+	struct dirent *dent;
+	struct stat st;
+	while (++count < DIRLIST_BATCH && (dent = readdir(p->dp)) != NULL)
+  #endif
+	{
+	  #ifdef _WIN32
+		/* WC_ERR_INVALID_CHARS not used in string conversion since
+		 * expecting valid unicode from reading directory */
+		const char * const d_name = p->fnUTF8;
+		uint32_t dsz = (uint32_t)
+		  WideCharToMultiByte(CP_UTF8, 0, p->ffd.cFileName, -1,
+		                      p->fnUTF8, sizeof(p->fnUTF8), NULL, NULL);
+		if (0 == dsz) continue;
+		--dsz;
+	  #else
 		const char * const d_name = dent->d_name;
 		const uint32_t dsz = (uint32_t) _D_EXACT_NAMLEN(dent);
+	  #endif
 		if (d_name[0] == '.') {
 			if (hide_dotfiles)
 				continue;
@@ -1081,6 +1153,7 @@ static int http_read_directory(handler_ctx * const p) {
 		force_assert(dsz < sizeof(dent->d_name));
 	  #endif
 
+	  #ifndef _WIN32
 	  #ifndef _ATFILE_SOURCE
 		memcpy(p->path_file, d_name, dsz + 1);
 		if (stat(p->path, &st) != 0)
@@ -1089,6 +1162,7 @@ static int http_read_directory(handler_ctx * const p) {
 		/*(XXX: follow symlinks, like stat(); not using AT_SYMLINK_NOFOLLOW) */
 		if (0 != fstatat(p->dfd, d_name, &st, 0))
 			continue; /* file *just* disappeared? */
+	  #endif
 	  #endif
 
 		if (p->jb) { /* json output */
@@ -1098,11 +1172,16 @@ static int http_read_directory(handler_ctx * const p) {
 				p->jcomma = 1;
 				buffer_append_string_len(p->jb, CONST_STR_LEN( "{\"name\":\""));
 			}
-			buffer_append_bs_escaped(p->jb, d_name, dsz, BS_ESCAPE_JSON);
+			buffer_append_bs_escaped_json(p->jb, d_name, dsz);
 
 			const char *t;
 			size_t tlen;
-			if (!S_ISDIR(st.st_mode)) {
+		  #ifndef _WIN32
+			if (!S_ISDIR(st.st_mode))
+		  #else
+			if (!(p->ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
+		  #endif
+			{
 				t =           "\",\"type\":\"file\",\"size\":";
 				tlen = sizeof("\",\"type\":\"file\",\"size\":")-1;
 			}
@@ -1114,33 +1193,69 @@ static int http_read_directory(handler_ctx * const p) {
 			char mstr[LI_ITOSTRING_LENGTH];
 			struct const_iovec iov[] = {
 			  { t, tlen }
+			 #ifndef _WIN32
 			 ,{ sstr, li_itostrn(sstr, sizeof(sstr), st.st_size) }
 			 ,{ CONST_STR_LEN(",\"mtime\":") }
 			 ,{ mstr, li_itostrn(mstr, sizeof(mstr), TIME64_CAST(st.st_mtime)) }
+			 #else
+			 ,{ sstr, li_itostrn(sstr, sizeof(sstr),
+			                     ((int64_t)p->ffd.nFileSizeHigh << 32)
+			                             | p->ffd.nFileSizeLow) }
+			 ,{ CONST_STR_LEN(",\"mtime\":") }
+			 ,{ mstr, li_itostrn(mstr, sizeof(mstr),
+			            ((((int64_t)p->ffd.ftLastWriteTime.dwHighDateTime << 32)
+			                      | p->ffd.ftLastWriteTime.dwLowDateTime)
+			            / 10000000 - 11644473600LL)) }
+			 #endif
 			 ,{ CONST_STR_LEN("}") }
 			};
 			buffer_append_iovec(p->jb, iov, sizeof(iov)/sizeof(*iov));
 			continue;
 		}
 
+	  #ifndef _WIN32
 		dirls_list_t * const list = !S_ISDIR(st.st_mode) ? &p->files : &p->dirs;
-		if (list->used == list->size) {
-			list->size += DIRLIST_BLOB_SIZE;
-			list->ent   = (dirls_entry_t**) realloc(list->ent, sizeof(dirls_entry_t*) * list->size);
-			force_assert(list->ent);
-		}
+	  #else  /* _WIN32 */
+		dirls_list_t * const list =
+		  !(p->ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
+		    ? &p->files
+		    : &p->dirs;
+	  #endif /* _WIN32 */
+		if (!(list->used & (DIRLIST_BLOB_SIZE-1)))
+			ck_realloc_u32((void **)&list->ent, list->used,
+			               DIRLIST_BLOB_SIZE, sizeof(*list->ent));
 		dirls_entry_t * const tmp = list->ent[list->used++] =
-		  (dirls_entry_t*) malloc(sizeof(dirls_entry_t) + 1 + dsz);
-		force_assert(tmp);
+		  (dirls_entry_t*) ck_malloc(sizeof(dirls_entry_t) + 1 + dsz);
+	  #ifdef _WIN32 /*(convert 100ns ticks since 1 Jan 1601 to unix time_t)*/
+		/*(future: preserve FILETIME here and use Windows fn to format time
+		 * below instead of localtime_r() and buffer_append_strftime())*/
+		tmp->mtime = (time_t)
+                  ((((int64_t)p->ffd.ftLastWriteTime.dwHighDateTime << 32)
+		            | p->ffd.ftLastWriteTime.dwLowDateTime)
+                  / 10000000 - 11644473600LL);
+		tmp->size = ((int64_t)p->ffd.nFileSizeHigh << 32) | p->ffd.nFileSizeLow;
+	  #else
 		tmp->mtime = st.st_mtime;
 		tmp->size  = st.st_size;
+	  #endif
 		tmp->namelen = dsz;
 		memcpy(DIRLIST_ENT_NAME(tmp), d_name, dsz + 1);
 	}
+  #ifdef _WIN32
+	  while (++count < DIRLIST_BATCH && FindNextFileW(p->hFind, &p->ffd) != 0);
+	if (count == DIRLIST_BATCH)
+		return HANDLER_WAIT_FOR_EVENT;
+	if (GetLastError() != ERROR_NO_MORE_FILES) {
+		/*(some other GetLastError() value; ignore; truncate listing)*/
+	}
+	FindClose(p->hFind);
+	p->hFind = INVALID_HANDLE_VALUE;
+  #else
 	if (count == DIRLIST_BATCH)
 		return HANDLER_WAIT_FOR_EVENT;
 	closedir(p->dp);
 	p->dp = NULL;
+  #endif
 
 	return HANDLER_FINISHED;
 }
@@ -1175,7 +1290,11 @@ static void http_list_directory(request_st * const r, handler_ctx * const hctx)
 		buffer_append_string_len(out, CONST_STR_LEN("/\">"));
 		buffer_append_string_encoded(out, DIRLIST_ENT_NAME(tmp), tmp->namelen, ENCODING_MINIMAL_XML);
 		buffer_append_string_len(out, CONST_STR_LEN("</a>/</td><td class=\"m\">"));
+	  #ifdef __MINGW32__
+		buffer_append_strftime(out, "%Y-%b-%d %H:%M:%S", localtime64_r(&tmp->mtime, &tm));
+	  #else
 		buffer_append_strftime(out, "%Y-%b-%d %T", localtime64_r(&tmp->mtime, &tm));
+	  #endif
 		buffer_append_string_len(out, CONST_STR_LEN("</td><td class=\"s\">- &nbsp;</td><td class=\"t\">Directory</td></tr>\n"));
 
 		if (buffer_string_space(out) < 256) {
@@ -1216,7 +1335,11 @@ static void http_list_directory(request_st * const r, handler_ctx * const hctx)
 		buffer_append_string_len(out, CONST_STR_LEN("\">"));
 		buffer_append_string_encoded(out, DIRLIST_ENT_NAME(tmp), tmp->namelen, ENCODING_MINIMAL_XML);
 		buffer_append_string_len(out, CONST_STR_LEN("</a></td><td class=\"m\">"));
+	  #ifdef __MINGW32__
+		buffer_append_strftime(out, "%Y-%b-%d %H:%M:%S", localtime64_r(&tmp->mtime, &tm));
+	  #else
 		buffer_append_strftime(out, "%Y-%b-%d %T", localtime64_r(&tmp->mtime, &tm));
+	  #endif
 		size_t buflen =
 		  http_list_directory_sizefmt(sizebuf, sizeof(sizebuf), tmp->size);
 		struct const_iovec iov[] = {
@@ -1285,11 +1408,11 @@ static void mod_dirlisting_json_append (request_st * const r, handler_ctx * cons
     if (hctx->jfn) {
         if (__builtin_expect( (write_all(hctx->jfd, BUF_PTR_LEN(jb)) < 0), 0)) {
             /*(cleanup, cease caching if error occurs writing to cache file)*/
+            close(hctx->jfd);
+            hctx->jfd = -1;
             unlink(hctx->jfn);
             free(hctx->jfn);
             hctx->jfn = NULL;
-            close(hctx->jfd);
-            hctx->jfd = -1;
         }
         /* Note: writing cache file is separate from the response so that if an
          * error occurs with cache, the response still proceeds.  While this is
@@ -1350,11 +1473,29 @@ URIHANDLER_FUNC(mod_dirlisting_subrequest_start) {
 	}
   #endif
 
-	/* TODO: add option/mechanism to enable json output */
   #if 0 /* XXX: ??? might this be enabled accidentally by clients ??? */
+	/* XXX: would have to add "Vary: Accept" response header, too */
 	const buffer * const vb =
 	  http_header_request_get(r, HTTP_HEADER_ACCEPT, CONST_STR_LEN("Accept"));
 	p->conf.json = (vb && strstr(vb->ptr, "application/json")); /*(coarse)*/
+	if (p->conf.json) p->conf.auto_layout = 0;
+  #else
+	/* check URL for /<path>/?json to enable json output */
+	if (buffer_clen(&r->uri.query) == sizeof("json")-1
+	    && 0 == memcmp(r->uri.query.ptr, CONST_STR_LEN("json")-1)) {
+	  #if 0
+		/* streaming response not set here for mod_deflate (which
+		 * currently does not compress incomlete streaming responses),
+		 * since json response is generally highly compressible.
+		 * Admin should enable streaming response in lighttpd.conf,
+		 * if desired. */
+		if (!(r->conf.stream_response_body
+		      & (FDEVENT_STREAM_RESPONSE|FDEVENT_STREAM_RESPONSE_BUFMIN)))
+			r->conf.stream_response_body |= FDEVENT_STREAM_RESPONSE;
+	  #endif
+		p->conf.json = 1;
+		p->conf.auto_layout = 0;
+	}
   #endif
 
 	if (p->conf.cache) {
@@ -1410,10 +1551,25 @@ URIHANDLER_FUNC(mod_dirlisting_subrequest_start) {
 		r->http_status = 200;
 		r->resp_body_started = 1;
 	}
+	else if (p->conf.auto_layout)
+		http_dirlist_auto_layout_early_hints(r, p);
 
 	r->plugin_ctx[p->id] = hctx;
 	r->handler_module = p->self;
-	return mod_dirlisting_subrequest(r, p);
+	handler_t rc = mod_dirlisting_subrequest(r, p);
+
+	if (rc == HANDLER_WAIT_FOR_EVENT && p->conf.auto_layout
+	    && (p->conf.external_js || p->conf.external_css)
+	      /* (paranoia: do not send 103 for HTTP/1.x; only for HTTP/2 +)
+	       *  https://www.rfc-editor.org/rfc/rfc8297.html#section-3 */
+	    && r->http_version >= HTTP_VERSION_2) {
+		/* send 103 Early Hints intermediate response (send once only)*/
+		r->http_status = 103; /* 103 Early Hints */
+		if (!http_response_send_1xx(r))
+			rc = HANDLER_ERROR;
+	}
+
+	return rc;
 }
 
 
@@ -1462,9 +1618,34 @@ REQUEST_FUNC(mod_dirlisting_reset) {
 }
 
 
+static void mod_dirlisting_cache_control (request_st * const r, unix_time64_t max_age) {
+    if (!light_btst(r->resp_htags, HTTP_HEADER_CACHE_CONTROL)) {
+        buffer * const vb =
+          http_header_response_set_ptr(r, HTTP_HEADER_CACHE_CONTROL,
+                                       CONST_STR_LEN("Cache-Control"));
+        buffer_append_string_len(vb, CONST_STR_LEN("max-age="));
+        buffer_append_int(vb, max_age);
+    }
+}
+
+
+static void mod_dirlisting_cache_etag (request_st * const r, int fd)
+{
+    if (0 != r->conf.etag_flags) {
+        struct stat st;
+        if (0 == fstat(fd, &st)) {
+            buffer * const vb =
+              http_header_response_set_ptr(r, HTTP_HEADER_ETAG,
+                                           CONST_STR_LEN("ETag"));
+            http_etag_create(vb, &st, r->conf.etag_flags);
+        }
+    }
+}
+
+
 static handler_t mod_dirlisting_cache_check (request_st * const r, plugin_data * const p) {
     /* optional: an external process can trigger a refresh by deleting the cache
-     * entry when the external process detects (or initiaties) changes to dir */
+     * entry when the external process detects (or initiates) changes to dir */
     buffer * const tb = r->tmp_buf;
     buffer_copy_path_len2(tb, BUF_PTR_LEN(p->conf.cache->path),
                               BUF_PTR_LEN(&r->physical.path));
@@ -1473,7 +1654,9 @@ static handler_t mod_dirlisting_cache_check (request_st * const r, plugin_data *
     stat_cache_entry * const sce = stat_cache_get_entry_open(tb, 1);
     if (NULL == sce || sce->fd == -1)
         return HANDLER_GO_ON;
-    if (TIME64_CAST(sce->st.st_mtime) + p->conf.cache->max_age < log_epoch_secs)
+    const unix_time64_t max_age =
+      TIME64_CAST(sce->st.st_mtime) + p->conf.cache->max_age - log_epoch_secs;
+    if (max_age < 0)
         return HANDLER_GO_ON;
 
     !p->conf.json
@@ -1505,6 +1688,18 @@ static handler_t mod_dirlisting_cache_check (request_st * const r, plugin_data *
         return HANDLER_GO_ON;
     }
 
+    /* Cache-Control and ETag (also done in mod_dirlisting_cache_add())*/
+    mod_dirlisting_cache_control(r, max_age);
+    if (0 != r->conf.etag_flags) {
+        const buffer *etag = stat_cache_etag_get(sce, r->conf.etag_flags);
+        if (etag && !buffer_is_blank(etag))
+            http_header_response_set(r, HTTP_HEADER_ETAG,
+                                     CONST_STR_LEN("ETag"),
+                                     BUF_PTR_LEN(etag));
+    }
+    if (p->conf.auto_layout)
+        http_dirlist_auto_layout_early_hints(r, p);
+
     r->resp_body_finished = 1;
     return HANDLER_FINISHED;
 }
@@ -1535,8 +1730,6 @@ static int mod_dirlisting_write_cq (const int fd, chunkqueue * const cq, log_err
 }
 
 
-#include <sys/stat.h>   /* mkdir() */
-#include <sys/types.h>
 /*(similar to mod_deflate.c:mkdir_recursive(), but starts mid-path)*/
 static int mkdir_recursive (char *dir, size_t off) {
     char *p = dir+off;
@@ -1558,7 +1751,6 @@ static int mkdir_recursive (char *dir, size_t off) {
 }
 
 
-#include <stdio.h>      /* rename() */
 __attribute_noinline__
 static void mod_dirlisting_cache_add (request_st * const r, handler_ctx * const hctx) {
   #ifndef PATH_MAX
@@ -1580,12 +1772,16 @@ static void mod_dirlisting_cache_add (request_st * const r, handler_ctx * const
     memcpy(oldpath, tb->ptr, len+7+1); /*(include '\0')*/
     const int fd = fdevent_mkostemp(oldpath, 0);
     if (fd < 0) return;
-    if (mod_dirlisting_write_cq(fd, &r->write_queue, r->conf.errh)
-        && 0 == rename(oldpath, newpath))
+    int rc = mod_dirlisting_write_cq(fd, &r->write_queue, r->conf.errh);
+    if (rc)
+        mod_dirlisting_cache_etag(r, fd);
+    close(fd); /*(_WIN32 rename fails if file open; MS filesystem limitation)*/
+    if (rc && 0 == fdevent_rename(oldpath, newpath)) {
         stat_cache_invalidate_entry(newpath, len);
+        mod_dirlisting_cache_control(r, hctx->conf.cache->max_age);
+    }
     else
         unlink(oldpath);
-    close(fd);
 }
 
 
@@ -1605,8 +1801,7 @@ static void mod_dirlisting_cache_json_init (request_st * const r, handler_ctx *
     if (fd < 0) return;
     hctx->jfn_len = buffer_clen(tb);
     hctx->jfd = fd;
-    hctx->jfn = malloc(hctx->jfn_len+1);
-    force_assert(hctx->jfn);
+    hctx->jfn = ck_malloc(hctx->jfn_len+1);
     memcpy(hctx->jfn, tb->ptr, hctx->jfn_len+1); /*(include '\0')*/
 }
 
@@ -1622,8 +1817,15 @@ static void mod_dirlisting_cache_json (request_st * const r, handler_ctx * const
     force_assert(len < PATH_MAX);
     memcpy(newpath, hctx->jfn, len);
     newpath[len] = '\0';
-    if (0 == rename(hctx->jfn, newpath))
+    if (0 == r->resp_header_len) /*(response headers not yet sent)*/
+        mod_dirlisting_cache_etag(r, hctx->jfd);
+    close(hctx->jfd); /*(_WIN32 rename fails if file open; MS fs limitation)*/
+    hctx->jfd = -1;
+    if (0 == fdevent_rename(hctx->jfn, newpath)) {
         stat_cache_invalidate_entry(newpath, len);
+        if (0 == r->resp_header_len) /*(response headers not yet sent)*/
+            mod_dirlisting_cache_control(r, hctx->conf.cache->max_age);
+    }
     else
         unlink(hctx->jfn);
     free(hctx->jfn);
@@ -1631,6 +1833,8 @@ static void mod_dirlisting_cache_json (request_st * const r, handler_ctx * const
 }
 
 
+__attribute_cold__
+__declspec_dllexport__
 int mod_dirlisting_plugin_init(plugin *p);
 int mod_dirlisting_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
